﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Manage.DataContracts;
using Manage.Services.Synchronizers;
using Newtonsoft.Json;
using Orchestration;
using UnityEngine;
using UnityEngine.SceneManagement;

namespace Manage.Services
{
    public class ManageService : MonoBehaviour, IStartupComponent
    {
        public static string serverUrl;
        public bool logMessages;
        public TaskSynchronizer taskSynchronizer;
        public PersonSynchronizer personSynchronizer;
        public GoalSynchronizer goalSynchronizer;
        public bool isLocalHost;

        private DdpConnection ddpConnection;
        private LocalDB localDB;
        private JsonObjectCollection goalCollection;
        private JsonObjectCollection taskCollection;
        private JsonObjectCollection groupCollection;
        private JsonObjectCollection personCollection;
        private static string localHost = "ws://localhost:3000/websocket";
        private static string production = "ws://m-nage-ole.eu.meteorapp.com/websocket";

        private Queue<MethodCall> methodQueue;

        public bool IsConnected { get; set; }

        #region IStartupComponent

        public int InitOrder { get { return 0; } }
        public bool IsInitComplete { get; private set; }

        public void Initialize()
        {
            methodQueue = new Queue<MethodCall>();
            SetupConnection();
            SetupDdpCallbacks();
            Connect();
            SetupCallbacks();
            StartCoroutine(MessageQueueProcessor());
        }

        #endregion IStartupComponent

        public void OnApplicationFocus(bool focus)
        {
            if (!focus || !IsInitComplete)
            {
                return;
            }

            if (ddpConnection != null)
            {
                if (ddpConnection.GetConnectionState() != DdpConnection.ConnectionState.CONNECTED)
                {
                    IsConnected = false;
                    Debug.LogWarning("OnApplicationFocus: connection was: {0}" + ddpConnection.GetConnectionState());
                    Connect(); // TODO: Fix error when reconnecting
                }
            }
            else
            {
                Debug.LogError("The connection was null.");
                SceneManager.LoadSceneAsync("MainGameplay");
            }
        }

        private IEnumerator MessageQueueProcessor()
        {
            while (true)
            {
                if (methodQueue.Count > 0)
                {
                    var methodCall = methodQueue.Dequeue();
                    ddpConnection.Call(methodCall.methodName, methodCall.items);
                }
                yield return null;
            }
        }

        private void SetupConnection()
        {
            serverUrl = serverUrl ?? ((isLocalHost) ? localHost : production);
            Application.runInBackground = true;
            // Initialize connection
            ddpConnection = new DdpConnection(serverUrl);
            ddpConnection.logMessages = logMessages;

            // Set up local DB
            localDB = new LocalDB((db, collectionName) => new JsonObjectCollection(db, collectionName), ddpConnection);
            groupCollection = (JsonObjectCollection)localDB.GetCollection("groups");
            taskCollection = (JsonObjectCollection)localDB.GetCollection("tasks");
            personCollection = (JsonObjectCollection)localDB.GetCollection("profiles");
            goalCollection = (JsonObjectCollection)localDB.GetCollection("goals");
        }

        private void SetupDdpCallbacks()
        {
            ddpConnection.OnConnected += (DdpConnection connection) =>
            {
                Debug.Log("Connected.");
                IsConnected = true;
                IsInitComplete = true;
            };

            ddpConnection.OnDisconnected += (DdpConnection connection) =>
            {
                Debug.LogWarning("Disconnected.");

                StartCoroutine(CoroutineHelper.GetInstance().RunAfter(() =>
                {
                    Debug.Log("Try to reconnect ...");
                    connection.Connect();
                }, 2.0f));
            };

            ddpConnection.OnError += (DdpError error) =>
            {
                Debug.LogError("Error: " + error.errorCode + " " + error.reason);
            };
        }

        private void SetupCallbacks()
        {
            SetupTaskCollection();
            SetupGroupCollection();
            SetupPersonCollection();
            SetupGoalCollection();
        }

        private void SetupGoalCollection()
        {
            goalCollection.OnAdded += (id, fields) =>
            {
                if (goalSynchronizer != null)
                {
                    var goalDto = ConvertTo<GoalDto>(new KeyValuePair<string, JSONObject>(id, fields));
                    StartCoroutine(goalSynchronizer.AddGoal(goalDto));
                }

            };

            goalCollection.OnRemoved += (id) =>
            {
                // TODO : something
            };

            goalCollection.OnChanged += (id, fields, cleared) =>
            {
                // TODO : something
            };
        }

        private void SetupTaskCollection()
        {
            taskCollection.OnAdded += (id, fields) =>
            {
                if (taskSynchronizer != null)
                {
                    var taskDto = ConvertTo<TaskDto>(new KeyValuePair<string, JSONObject>(id, fields));
                    StartCoroutine(taskSynchronizer.AddNewTask(taskDto));
                }
            };

            taskCollection.OnRemoved += (id) => { taskSynchronizer.RemoveTask(id); };

            taskCollection.OnChanged += (id, fields, cleared) =>
            {
                if (taskSynchronizer != null)
                    taskSynchronizer.UpdateTask(ConvertTo<TaskDto>(new KeyValuePair<string, JSONObject>(id, fields)));
            };
        }

        private void SetupGroupCollection()
        {
            groupCollection.OnAdded += (id, fields) =>
            {
                personSynchronizer.AddFamily(ConvertTo<FamilyDto>(new KeyValuePair<string, JSONObject>(id, fields)));
            };

            groupCollection.OnRemoved += (id) => { };

            groupCollection.OnChanged += (id, fields, cleared) => { };
        }

        private void SetupPersonCollection()
        {
            personCollection.OnAdded += (id, fields) =>
            {
                if (personSynchronizer != null)
                    personSynchronizer.AddPerson(ConvertTo<PersonDto>(new KeyValuePair<string, JSONObject>(id, fields)));
            };

            personCollection.OnRemoved += (id) => { };

            personCollection.OnChanged += (id, fields, cleared) =>
            {
                if (personSynchronizer != null)
                    personSynchronizer.PersonChanged(ConvertTo<PersonDto>(new KeyValuePair<string, JSONObject>(id, fields)));
            };
        }

        public Subscription Subscribe(string subscriptionName)
        {
            var subscription = ddpConnection.Subscribe(subscriptionName);
            subscription.OnReady = (Subscription obj) =>
            {
                Debug.Log("Ready subscription: " + obj.id);
            };
            return subscription;
        }

        public void UnSubscribe(Subscription subscription)
        {
            ddpConnection.Unsubscribe(subscription);
        }

        private void Connect()
        {
            Debug.Log("Connecting ...");
            ddpConnection.Connect();
        }

        public void CallMethod(string methodName, params JSONObject[] items)
        {
            var methodCall = new MethodCall
            {
                methodName = methodName,
                items = items
            };
            methodQueue.Enqueue(methodCall);
        }

        public static T ConvertTo<T>(KeyValuePair<string, JSONObject> entry) where T : IHaveId
        {
            var deserializeObject = JsonConvert.DeserializeObject<T>(entry.Value.ToString());
            deserializeObject.id = entry.Key;

            return deserializeObject;
        }

        public static string GetServerUrlHttp()
        {
            return serverUrl.Replace("ws://", "http://").Replace("websocket", "");
        }
    }
}
