﻿using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Data;
using Manage.DataContracts;
using Manage.Services.Contracts;
using Orchestration;
using UnityEngine;

namespace Manage.Services.Synchronizers
{
    public class GoalSynchronizer : MonoBehaviour, IGoalSynchronizer, IStartupComponent
    {
        public PoolController poolController;
        public ManageService manageService;
        public PersonSynchronizer personSynchrinizer;

        public List<Goal> Goals { get; private set; }

        private Subscription _goalSubscription;
        private Subscription _goalInstancesSubscription;

        public GoalSynchronizer()
        {
            Goals = new List<Goal>();
        }

        #region IStartupComponent

        public int InitOrder { get { return 5; } }
        public bool IsInitComplete { get; private set; }

        public void Initialize()
        {
            _goalSubscription = manageService.Subscribe("goals");
            _goalInstancesSubscription = manageService.Subscribe("goalInstances");

            StartCoroutine(WaitForInit());
        }

        private IEnumerator WaitForInit()
        {
            while (!_goalSubscription.isReady || Goals.Count < 1 || !_goalInstancesSubscription.isReady)
            {
                yield return null;
            }
            IsInitComplete = true;
        }

        #endregion

        public IEnumerator AddGoal(GoalDto goalDto)
        {
            var goal = BuildGoal(goalDto);
            var isNotComplete = !goal.IsCompleted.HasValue || !goal.IsCompleted.Value;
            if (goal.Person != null)
            {
                GameObject rewardPool;
                while ((rewardPool = poolController.GetRewardPool(goal.Person)) == null)
                    yield return null;
                rewardPool.GetComponent<RewardPool>().AddGoal(goal);

            }
            else if (goal.IsGroup.HasValue && goal.IsGroup.Value
                     && goal.IsAssigned.HasValue && goal.IsAssigned.Value)
            {
                poolController.FamilyPool.AddGoal(goal);
            }
            else if (isNotComplete)
            {
                Goals.Add(goal);
            }
        }

        private Goal BuildGoal(GoalDto goalDto)
        {
            var goalSprite = GameLogic.Instance.goalSprites[goalDto.iconIndex.Value];
            var goal = new Goal
            {
                Id = goalDto.id,
                Description = goalDto.name,
                Sprite = goalSprite,
                IsCompleted = goalDto.isCompleted,
                IsGroup = goalDto.isGroup,
                Score = goalDto.score,
                IsAssigned = goalDto.isAssigned
            };
            if (!string.IsNullOrEmpty(goalDto.personId))
                goal.Person = personSynchrinizer.Persons.First(p => p.Id == goalDto.personId);
            return goal;
        }

        #region IGoalSynchronizer

        public void AssignGoal(Person person, Goal goal)
        {
            if (person != null)
            {
                manageService.CallMethod("goals.assignTo",
                    JSONObject.StringObject(goal.Id),
                    JSONObject.StringObject(person.Id));
            }
            else
            {
                manageService.CallMethod("goals.assignToGroup",
                    JSONObject.StringObject(goal.Id),
                    JSONObject.StringObject(personSynchrinizer.Family.Id));
            }
        }

        public void CompleteGoal(Goal goal, List<TaskBallController> taskControllers)
        {
            // Remove tasks from other pool
            if (goal.Person != null)
            {
                foreach (var t in taskControllers)
                {
                    poolController.FamilyPool.RemoveTask(t.gameObject);
                }
            }
            else
            {
                foreach (var taskController in taskControllers)
                {
                    var person = taskController.Task.Person;
                    var rewardPool = poolController.GetRewardPool(person).GetComponent<RewardPool>();
                    rewardPool.RemoveTask(taskController.gameObject);
                }
            }

            var taskIdArray = JSONObject.Create(JSONObject.Type.ARRAY);
            foreach (var t in taskControllers)
            {
                taskIdArray.Add(t.Task.Id);
            }
            manageService.CallMethod("goals.isCompleted", JSONObject.StringObject(goal.Id), taskIdArray);
        }

        public void RedeemGoal(Goal goal)
        {
            var rewardPool = (goal.Person != null)
                ? poolController.GetRewardPool(goal.Person).GetComponent<RewardPool>()
                : poolController.FamilyPool;
            rewardPool.GoalPool.RemoveGoal(goal);
            manageService.CallMethod("goals.isRedeemed", JSONObject.StringObject(goal.Id));
        }

        #endregion IGoalSynchronizer
    }
}
