﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Constants;
using Data;
using Data.Enums;
using Manage.DataContracts;
using Manage.Services.Contracts;
using Orchestration;
using UnityEngine;

namespace Manage.Services.Synchronizers
{
    public class TaskSynchronizer : MonoBehaviour, ITaskSynchronizer, IStartupComponent
    {
        public GameObject taskPrefab;
        public GameObject photoTaskPrefab;
        public TaskBarUI taskBarUi;
        public PoolController poolController;
        public ManageService manageService;
        public PersonSynchronizer personSynchronizer;

        private Subscription _taskSubscription;
        private Subscription _taskInstancesSubscription;
        private Dictionary<string, Sprite> taskIconCache = new Dictionary<string, Sprite>();

        public List<Task> Tasks { get; private set; }

        public TaskSynchronizer()
        {
            Tasks = new List<Task>();
        }

        #region IStartupComponent

        public int InitOrder { get { return 7; } }
        public bool IsInitComplete { get; private set; }

        public void Initialize()
        {
            _taskSubscription = manageService.Subscribe("tasks");
            _taskInstancesSubscription = manageService.Subscribe("taskInstances");

            StartCoroutine(WaitForInit());
        }

        private IEnumerator WaitForInit()
        {
            while (!_taskSubscription.isReady || !_taskInstancesSubscription.isReady)
            {
                yield return null;
            }
            IsInitComplete = true;
        }

        #endregion

        public void OnDestroy()
        {
            if (manageService != null)
            {
                if (_taskSubscription != null)
                    manageService.UnSubscribe(_taskSubscription);
                if (_taskInstancesSubscription != null)
                    manageService.UnSubscribe(_taskInstancesSubscription);
            }

        }

        #region ITaskSynchronizer
        public void DeleteTask(Task task)
        {
            manageService.CallMethod("tasks.remove", JSONObject.StringObject(task.Id));
        }
        #endregion ITaskSynchronizer

        private Task BuildTask(TaskDto dtoTask, Person person = null)
        {
            Task task = new Task();
            return BuildTask(dtoTask, ref task);
        }

        private Task BuildTask(TaskDto dtoTask, ref Task task, Person person = null)
        {
            task.IsUpdateEnabled = false;
            task.Id = dtoTask.id;
            task.IsDirty = false;
            if (dtoTask.taskIconUrl != null)
            {
                task.TaskIcon = taskIconCache[dtoTask.taskIconUrl];
            }
            else if (dtoTask.taskIconIndex.HasValue)
            {
                var index = dtoTask.taskIconIndex.Value;
                if (index < 0 || index >= GameLogic.Instance.taskIcons.Count) index = 0;
                task.TaskIcon = GameLogic.Instance.taskIcons[index];
            }
            if (!string.IsNullOrEmpty(dtoTask.personId))
                task.Person = personSynchronizer.Persons.FirstOrDefault(x => x.Id == dtoTask.personId);
            if (dtoTask.score.HasValue)
                task.Score = dtoTask.score ?? 0;
            if (dtoTask.name != null)
                task.TaskName = dtoTask.name;
            if (dtoTask.age.HasValue)
                task.Age = (AgeGroup)dtoTask.age;
            if (dtoTask.isCompleted.HasValue)
                task.IsCompleted = dtoTask.isCompleted.Value;
            if (dtoTask.isGroupTask.HasValue)
                task.IsGroupTask = dtoTask.isGroupTask.Value;
            if (dtoTask.isGoaled.HasValue)
                task.IsGoaled = dtoTask.isGoaled.Value;
            if (dtoTask.isApproved.HasValue)
                task.IsApproved = dtoTask.isApproved.Value;
            task.IsUpdateEnabled = true;
            return task;
        }

        public GameObject CreateTaskGameObject(Task task)
        {
            Tasks.Add(task);
            task.PropertyChanged += Task_PropertyChanged;
            GameObject go = (!taskIconCache.ContainsValue(task.TaskIcon))
                ? Instantiate(taskPrefab) : Instantiate(photoTaskPrefab);
            var taskBallController = go.GetComponent<TaskBallController>();
            taskBallController.Task = task;
            taskBallController.taskBarUi = taskBarUi;
            taskBallController.transform.parent = TaskBallController.GetTaskRoot();
            taskBallController.TaskSynchronizer = this;
            go.GetComponent<TaskView>().Task = task;
            float scale = Mathf.Min(task.Score, 10) / 20f + 0.6f; // TODO: fix hardcode.
            // Quick fix for growing task: TODO: fix this
            if (task.IsApproved)
                scale = 1.05f;
            go.transform.localScale = new Vector3(scale, scale, 1);
            go.transform.position = new Vector2(-100, -100);
            return go;
        }

        private void SpawnTasksInPool(List<GameObject> tasks, UnityEngine.Bounds poolBounds)
        {
            float x = poolBounds.min.x;
            float y = poolBounds.min.y;
            float nextY = y;
            foreach (var task in tasks)
            {
                float radius = task.GetComponent<CircleCollider2D>().bounds.extents.x;
                float diameter = radius * 2;
                if (y + diameter * 1.2f > poolBounds.max.y)
                    break;
                if (x + diameter > poolBounds.max.x)
                {
                    y = nextY;
                    x = poolBounds.min.x;
                }
                task.transform.position = new Vector2(x + radius, y + radius);
                task.GetComponent<TaskBallController>().EnablePhysics();
                x += diameter;
                nextY = Mathf.Max(nextY, y + diameter);
            }
        }

        public void RemoveTask(Task task)
        {
            Tasks.Remove(task);
            task.PropertyChanged -= Task_PropertyChanged;
        }

        public IEnumerator AddNewTask(TaskDto taskDto)
        {
            if (taskDto.clientId != null)
            {
                var existingTask = Tasks.FirstOrDefault(x => x.ClientId == taskDto.clientId);
                if (existingTask != null)
                {
                    existingTask.Id = taskDto.id;
                    existingTask.ClientId = null;
                    yield break;
                }
            }
            if (taskDto.taskIconUrl != null && !taskIconCache.ContainsKey(taskDto.taskIconUrl))
            {
                yield return StartCoroutine(FetchIconFromCache(taskDto.taskIconUrl));
            }
            var task = BuildTask(taskDto);
            GameObject go = CreateTaskGameObject(task);
            if (task.Person == null)
            {
                taskBarUi.AddTask(go);
            }
            else if (!task.IsCompleted)
            {
                var pool = poolController.GetTodoPool(task.Person).GetComponent<TodoPool>();
                pool.AddTask(go);
            }
            else
            {
                poolController.FamilyPool.AddTask(go);
                if (!task.IsGoaled)
                    poolController.GetRewardPool(task.Person).GetComponent<RewardPool>().AddTask(go);
            }
        }

        private IEnumerator FetchIconFromCache(string taskIconUrl)
        {
            if (taskIconCache.ContainsKey(taskIconUrl))
            {
                while (taskIconCache[taskIconUrl] == null)
                    yield return null;
            }
            else
            {
                yield return StartCoroutine(FetchIconFromUrl(taskIconUrl));
            }
        }

        private IEnumerator FetchIconFromUrl(string taskUrl)
        {
            taskIconCache.Add(taskUrl, null);
            var url = ManageService.GetServerUrlHttp() + taskUrl;
            Texture2D tex;
            tex = new Texture2D(4, 4, TextureFormat.DXT1, false);
            WWW www = new WWW(url);
            yield return www;
            www.LoadImageIntoTexture(tex);
            var rect = new Rect(0, 0, tex.width, tex.height);
            var sprite = Sprite.Create(tex, rect, new Vector2(0.5f, 0.5f));
            taskIconCache[taskUrl] = sprite;
        }

        private void Task_PropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
        {
            var task = (Task)sender;
            if (e.PropertyName == TaskProperties.IsCompleted)
            {
                manageService.CallMethod("tasks.isCompleted",
                    JSONObject.StringObject(task.Id),
                    JSONObject.Create(task.IsCompleted));

                task.Person.Experience += task.Score;
            }
            else if (e.PropertyName == TaskProperties.HasSettled && task.HasSettled)
            {
                var clientId = Guid.NewGuid().ToString();
                task.ClientId = clientId;
                manageService.CallMethod("tasks.assignTo",
                    JSONObject.StringObject(task.Id),
                    JSONObject.StringObject(task.Person.Id),
                    JSONObject.StringObject(clientId));
            }
            else if (e.PropertyName == TaskProperties.IsApproved && task.IsApproved)
            {
                manageService.CallMethod("tasks.isApproved",
                    JSONObject.StringObject(task.Id));

                task.Person.Experience += (task.Score / 5) * 2;
            }
        }

        public void UpdateTask(TaskDto taskDto)
        {
            var taskControllers = FindObjectsOfType<TaskBallController>();
            var taskController = taskControllers.First(x => x.Task.Id == taskDto.id);
            var task = taskController.Task;
            if (task != null)
            {
                var taskGo = taskController.gameObject;
                var person = task.Person;
                var isCompleted = task.IsCompleted;
                var isApproved = task.IsApproved;
                BuildTask(taskDto, ref task);
                if (taskDto.isApproved.HasValue)
                {
                    if (isApproved != task.IsApproved)
                        StartCoroutine(taskController.GrowTask());
                    else return;
                }
                if (!task.IsCompleted && (person != task.Person || isCompleted != task.IsCompleted))
                {
                    if (task.Person != null)
                    {
                        poolController.GetTodoPool(task.Person).GetComponent<TodoPool>().AddTask(taskGo);
                        if (person == null)
                            taskBarUi.Remove(taskGo);
                    }
                    else
                    {
                        taskController.DisablePhysics();
                        taskBarUi.AddTask(taskGo);
                    }
                }
                else if (task.IsCompleted)
                {
                    poolController.FamilyPool.AddTask(taskGo);
                    if (!task.IsGoaled && task.Person != null)
                        poolController.GetRewardPool(task.Person).GetComponent<RewardPool>().AddTask(taskGo);
                    if (person == null)
                        taskBarUi.Remove(taskGo);
                }
            }
        }

        public void RemoveTask(string id)
        {
            var taskController = FindObjectsOfType<TaskBallController>().First(x => x.Task.Id == id);
            var task = taskController.Task;
            if (task.IsGoaled)
                return;
            RemoveTask(taskController.gameObject);
        }

        public void RemoveTask(GameObject taskGo)
        {
            var taskController = taskGo.GetComponent<TaskBallController>();
            var task = taskController.Task;
            if (task.IsGoaled)
            {
                poolController.FamilyPool.RemoveTask(taskGo);
                if (task.Person != null)
                {
                    var rewardPool = poolController.GetRewardPool(task.Person).GetComponent<RewardPool>();
                    rewardPool.RemoveTask(taskGo);
                }
            }
            if (task.Person == null)
                taskBarUi.Remove(taskController.gameObject);
            Tasks.Remove(task);
            task.PropertyChanged -= Task_PropertyChanged;
            Destroy(taskController.gameObject);
        }
    }
}
