﻿using System;
using System.Collections;
using System.ComponentModel;
using Constants;
using Data;
using Manage.Services.Synchronizers;
using UnityEngine;
using UnityEngine.EventSystems;
using Camera = UnityEngine.Camera;

[RequireComponent(typeof(Rigidbody2D))]
[RequireComponent(typeof(CircleCollider2D))]
[RequireComponent(typeof(TaskView))]
public class TaskBallController : MonoBehaviour, IPointerDownHandler, IPointerUpHandler, IDragHandler
{
    public Rigidbody2D rigidBody;
    public new CircleCollider2D collider;
    public TaskView taskView;
    public TaskBarUI taskBarUi;

    public GameObject taskMenuUnassignedPrefab;
    public GameObject taskMenuAssignedPrefab;
    public GameObject taskMenuCompletedPrefab;

    public AudioSource hitBall;
    public AudioSource hitOther;
    public AudioSource release;
    public AudioSource drag;

    private Task task;
    private TaskAim taskAim;

    private Vector2 pivot;
    private Vector2 grabPoint;
    private bool isPressed;
    private bool isDragged;
    private bool isLaunched;
    private Person originalPerson;

    private GameObject dummyTask;

    public TaskSynchronizer TaskSynchronizer { get; set; }

    public bool IsDragged { get { return isDragged; } }

    public Vector2 Pivot
    {
        get { return pivot; }
        set { pivot = value; }
    }

    public Task Task
    {
        get { return task; }
        set
        {
            if (value == null) throw new ArgumentNullException();
            task = value;
            task.PropertyChanged += TaskPropertyChanged;
        }
    }

    private void TaskPropertyChanged(object sender, PropertyChangedEventArgs e)
    {
        if (e.PropertyName == TaskProperties.IsApproved && task.IsApproved)
        {
            StartCoroutine(GrowTask());
        }
    }

    public IEnumerator GrowTask()
    {
        // TODO: fix constants
        yield return new WaitForSeconds(0.2f);
        TaskSynchronizer.poolController.GetRewardPool(task.Person).GetComponent<RewardPool>().UpdatePoolLevel(task.Score * 0.4f);
        TaskSynchronizer.poolController.FamilyPool.UpdatePoolLevel(task.Score * 0.4f);
        while (transform.localScale.x < 1.05f)
        {
            transform.localScale += new Vector3(0.01f, 0.01f, 0f);
            yield return null;
        }
    }

    public void OnCollisionEnter2D(Collision2D collision)
    {
        if (task != null && !task.HasSettled && collision.relativeVelocity.magnitude > 0.1f)
        {
            if (collision.gameObject.GetComponent<TaskBallController>() != null)
            {
                hitBall.Play();
            }
            else
            {
                hitOther.Play();
            }
        }
    }

    public void Update()
    {
        if (isLaunched && task != null && !task.HasSettled)
        {
            if (rigidBody.velocity.magnitude < 0.1)
            { // Task is no longer moving
                if (task.Person != null) // Task was assigned to a person correctly
                    task.HasSettled = true;
                else
                    Destroy(gameObject);
                ReplaceWithDummy();
            }
            if (!Bounds.BoundingRectangle.Contains(transform.position))
            { // Task is out of bounds -> respawn at taskbar
                Destroy(gameObject);
                ReplaceWithDummy();
            }
        }
    }

    private void ReplaceWithDummy()
    {
        var dummyController = dummyTask.GetComponent<TaskBallController>();
        dummyController.SetAlpha(1);
        dummyController.collider.enabled = true;
    }

    public void Press()
    {
        if (!isPressed && !isDragged)
        {
            isPressed = true;
            grabPoint = Camera.main.ScreenToWorldPoint(Input.mousePosition);

            var assignee = task.Person != null ? task.Person.Name : "NULL";
            Debug.LogFormat("TaskId: {0}, name: {1}, isCompleted: {2}, isGroup: {3}, hasSettled: {4}, assignee: {5}",
                task.Id, task.TaskName, task.IsCompleted, task.IsGroupTask, task.HasSettled, assignee);
        }
    }

    public void Drag(Vector2 position)
    {
        if (!isPressed || (isLaunched && (!task.HasSettled || task.IsCompleted)))
            return;
        if (!isDragged)
        {
            if (Vector2.Distance(grabPoint, position) > Constants.TaskPhysics.dragThreshold)
            {
                isDragged = true;
                pivot = gameObject.transform.position;
                rigidBody.isKinematic = true;
                if (!task.HasSettled)
                {
                    GetComponent<TaskView>().SetSortingLayer(SortingLayers.TaskForeground);
                    taskAim = new GameObject("TaskAim", typeof(TaskAim)).GetComponent<TaskAim>();
                    taskAim.StartPosition = pivot;
                    drag.Play();
                    CreateDummyTask();
                }
                else
                {
                    collider.isTrigger = true;
                    originalPerson = task.Person;
                }
            }
        }
        if (isDragged)
        {
            if (!task.HasSettled)
            {
                // Constrain the task ball's movement to the defined radius
                bool aim = (Vector2.Distance(position, pivot) > Constants.TaskPhysics.safeRadius &&
                    Task.Person == null);
                SetEnableAim(aim);
                taskAim.UpdateAngle(position);
            }
            else if (task.Person == null &&
                (TaskSynchronizer.poolController.FamilyPool.IsInPool(collider) ||
                TaskSynchronizer.poolController.GetRewardPool(originalPerson)
                .GetComponent<RewardPool>().IsInPool(collider)))
            {
                task.Person = originalPerson;
            }
            transform.position = position;
        }
    }

    private void CreateDummyTask()
    {
        var newTask = new Task(task);
        dummyTask = TaskSynchronizer.CreateTaskGameObject(newTask);
        dummyTask.transform.SetParent(transform.parent, true);
        dummyTask.transform.position = pivot;
        var dummyController = dummyTask.GetComponent<TaskBallController>();
        dummyController.task.IsUpdateEnabled = true;
        dummyController.SetAlpha(0.3f);
        dummyController.collider.enabled = false;
    }

    public void SetAlpha(float alpha)
    {
        var spriteRenderers = gameObject.GetComponentsInChildren<SpriteRenderer>();
        foreach (var renderer in spriteRenderers)
        {
            var color = renderer.color;
            color.a = alpha;
            renderer.color = color;
        }
    }

    public void Release()
    {
        drag.Stop();
        if (isPressed && !isDragged)
        {
            Select();
        }
        else if (isDragged && !isLaunched)
        {
            isDragged = false;
            if (Vector2.Distance(transform.position, pivot) > Constants.TaskPhysics.safeRadius)
            {
                EnablePhysics();
                transform.SetParent(GetTaskRoot());
                taskBarUi.Replace(gameObject, dummyTask);
                taskView.SetSortingLayer(SortingLayers.TaskMiddle);
                Destroy(taskAim.gameObject);
                if (task.Person == null)
                    LaunchGameObject();
            }
            else
            {
                transform.position = pivot;
                Destroy(dummyTask);
            }
        }
        else if (isDragged && task.HasSettled)
        {
            isDragged = false;
            EnablePhysics();
            if (task.Person != null && task.Person != originalPerson)
            {
                // Cause an update by toggling the value off and on
                task.HasSettled = false;
                task.HasSettled = true;
            }
            else if (TaskSynchronizer.poolController.FamilyPool.IsInPool(collider)
                || TaskSynchronizer.poolController.GetRewardPool(originalPerson).GetComponent<RewardPool>().IsInPool(collider))
            {
                task.IsCompleted = true;
            }
            else
            {
                transform.position = pivot;
            }
            originalPerson = null;
        }
        isPressed = false;
    }

    public static Transform GetTaskRoot()
    {
        var taskRoot = GameObject.Find("TaskRoot");
        return taskRoot == null ? null : taskRoot.GetComponent<Transform>();
    }

    public void EnablePhysics()
    {
        isLaunched = true;
        rigidBody.isKinematic = false;
        collider.isTrigger = false;
    }

    public void DisablePhysics()
    {
        isLaunched = false;
        rigidBody.isKinematic = true;
        collider.isTrigger = true;
        transform.localRotation = Quaternion.identity;
    }

    public void Select()
    {
        if (isLaunched && !task.HasSettled)
        {
            return;
        }

        GameObject menuPrefab;
        if (task.Person != null)
        {
            menuPrefab = task.IsCompleted ? taskMenuCompletedPrefab : taskMenuAssignedPrefab;
        }
        else
        {
            menuPrefab = taskMenuUnassignedPrefab;
        }

        var canvas = GameObject.Find("Canvas").GetComponent<Canvas>();
        var taskMenu = Instantiate(menuPrefab).GetComponent<TaskMenu>();
        taskMenu.gameObject.transform.SetParent(canvas.transform, false);
        taskMenu.Task = task;
        taskMenu.TaskSynchronizer = TaskSynchronizer;
    }

    private void LaunchGameObject()
    {
        release.Play();
        Vector2 launchVector = pivot - (Vector2)transform.position;
        float magnitude = Mathf.Min(launchVector.magnitude / Constants.TaskPhysics.maxVelocityRadius, 1);
        magnitude *= Constants.TaskPhysics.maxLaunchVelocity;
        launchVector.Normalize();
        launchVector *= magnitude;
        rigidBody.velocity += launchVector;
        task.IsCompleted = false;
    }

    public void SetEnableAim(bool isEnabled)
    {
        if (taskAim != null && isEnabled != taskAim.gameObject.activeInHierarchy)
            taskAim.gameObject.SetActive(isEnabled);
    }

    public void OnDrag(PointerEventData eventData)
    {
        Vector2 dragPosition = Camera.main.ScreenToWorldPoint(Input.mousePosition);
        Drag(dragPosition);
    }

    public void OnPointerUp(PointerEventData eventData)
    {
        Release();
    }

    public void OnPointerDown(PointerEventData eventData)
    {
        Press();
    }
}
