﻿using System;
using System.Collections.Generic;
using Constants;
using Data;
using Manage.Services.Synchronizers;
using Orchestration;
using UnityEngine;
using UnityEngine.UI;

public class FamilyCircle : MonoBehaviour, IStartupComponent
{
    public float radius = 2;
    public int nPoints = 100;
    public float width = 0.08f;
    public Text familyName;
    public PersonSynchronizer personSynchronizer;

    private LineRenderer[] lines;
    private List<Person> persons;

    #region IStartupComponent

    public int InitOrder { get { return 8; } }
    public bool IsInitComplete { get; private set; }

    public void Initialize()
    {
        familyName.text = personSynchronizer.Family.Name;

        var collider = gameObject.AddComponent<CircleCollider2D>();
        collider.radius = radius;

        persons = personSynchronizer.Persons;
        if (persons.Count == 0)
        {
            GenerateLineSegment(Mathf.PI, Mathf.PI * 2, nPoints, Color.black);
        }
        else
        {
            foreach (var p in persons)
            {
                p.PropertyChanged += PersonPropertyChanged;
            }
            lines = new LineRenderer[persons.Count];
            float angleStart = Mathf.PI;
            float angleEnd = Mathf.PI * 2;
            float dAngle = (angleEnd - angleStart) / persons.Count;
            int nPointsPerson = nPoints / persons.Count + 1;
            int i;
            for (i = 0; i < persons.Count - 1; i++)
            {
                lines[i] = GenerateLineSegment(angleStart, angleStart += dAngle, nPointsPerson, persons[i].Color);
            }
            // Not in the loop because of floating point rounding errors
            lines[i] = GenerateLineSegment(angleStart, angleEnd, nPointsPerson, persons[i].Color);
        }

        IsInitComplete = true;
    }
    #endregion

    public void OnDestroy()
    {
        if (persons != null)
        {
            foreach (var p in persons)
            {
                p.PropertyChanged -= PersonPropertyChanged;
            }
        }
    }

    private void PersonPropertyChanged(object sender, System.ComponentModel.PropertyChangedEventArgs e)
    {
        if (e.PropertyName == PersonProperties.Color)
        {
            var person = (Person)sender;
            var index = persons.IndexOf(person);
            lines[index].material.SetColor("_EmissionColor", person.Color);
        }
    }

    private LineRenderer GenerateLineSegment(float angleStart, float angleEnd, int nPoints, Color color)
    {
        var go = new GameObject("SemiCircle", typeof(LineRenderer));
        var line = go.GetComponent<LineRenderer>();
        go.transform.SetParent(transform, false);
        line.material.SetColor("_EmissionColor", color);
        line.material.EnableKeyword("_EMISSION");
        line.useWorldSpace = false;
        line.startWidth = width;
        line.endWidth = width;
        line.numPositions = nPoints;
        line.SetPositions(GenerateSemiCircle(angleStart, angleEnd, nPoints));
        return line;
    }

    public Vector3[] GenerateSemiCircle(float angleStart, float angleEnd, int nPoints)
    {
        var points = new Vector3[nPoints];
        var point = 0;
        var dAngle = (angleEnd - angleStart) / (nPoints - 1);
        for (var angle = angleStart; angle < angleEnd; angle += dAngle)
        {
            points[point++] = new Vector2(Mathf.Cos(angle) * radius, Mathf.Sin(angle) * radius);
        }
        // Not in the loop because of floating point rounding errors
        points[nPoints - 1] = new Vector2(Mathf.Cos(angleEnd) * radius, Mathf.Sin(angleEnd) * radius);
        return points;
    }
}
